@@ -1,5 +1,7 @@
 import sys
+import typing
 from typing import (
+    AsyncIterator,
     Dict,
     Iterable,
     List,
@@ -17,13 +19,11 @@ from grpc import StatusCode
 from grpc.experimental.aio import AioRpcError
 from izihawa_utils.pb_to_json import ParseDict
 
-from .proto import beacon_service_pb2 as beacon_service_pb
 from .proto import consumer_service_pb2 as consumer_service_pb
 from .proto import index_service_pb2 as index_service_pb
 from .proto import query_pb2 as query_pb
 from .proto import reflection_service_pb2 as reflection_service_pb
 from .proto import search_service_pb2 as search_service_pb
-from .proto.beacon_service_pb2_grpc import BeaconApiStub
 from .proto.consumer_service_pb2_grpc import ConsumerApiStub
 from .proto.index_service_pb2_grpc import IndexApiStub
 from .proto.reflection_service_pb2_grpc import ReflectionApiStub
@@ -34,9 +34,17 @@ from .proto.utils_pb2 import (  # noqa
 )
 
 
+def setup_metadata(session_id, request_id):
+    metadata = []
+    if session_id:
+        metadata.append(('session-id', session_id))
+    if request_id:
+        metadata.append(('request-id', request_id))
+    return metadata
+
+
 class SummaClient(BaseGrpcClient):
     stub_clses = {
-        'beacon_api': BeaconApiStub,
         'consumer_api': ConsumerApiStub,
         'index_api': IndexApiStub,
         'reflection_api': ReflectionApiStub,
@@ -47,9 +55,8 @@ class SummaClient(BaseGrpcClient):
     async def attach_index(
         self,
         index_name: str,
-        attach_file_engine_request: Optional[Union[index_service_pb.AttachFileEngineRequest, Dict]] = None,
-        attach_remote_engine_request: Optional[Union[index_service_pb.AttachRemoteEngineRequest, Dict]] = None,
-        attach_ipfs_engine_request: Optional[Union[index_service_pb.AttachIpfsEngineRequest, Dict]] = None,
+        index_engine: dict,
+        merge_policy: Optional[Dict] = None,
         request_id: Optional[str] = None,
         session_id: Optional[str] = None,
     ) -> index_service_pb.AttachIndexResponse:
@@ -58,26 +65,24 @@ class SummaClient(BaseGrpcClient):
 
         Args:
             index_name: index name
-            attach_file_engine_request: attaching index placed under `<data_path>/<index_name>` directory
-            attach_remote_engine_request: attaching remote index
-            attach_ipfs_engine_request: attaching ipfs index
+            index_engine: {"file": {}}, {"memory": {}} or {"remote": {"chunked_cache_config": {"chunk_size": ..., "cache_size": ...}}}
+            merge_policy: describes how to select segments for merging, possible values: {"log": {}}, {"temporal": {}}
             request_id: request id
             session_id: session id
         """
         return await self.stubs['index_api'].attach_index(
             index_service_pb.AttachIndexRequest(
                 index_name=index_name,
-                attach_file_engine_request=attach_file_engine_request,
-                attach_remote_engine_request=attach_remote_engine_request,
-                attach_ipfs_engine_request=attach_ipfs_engine_request,
+                merge_policy=index_service_pb.MergePolicy(**(merge_policy or {})),
+                **index_engine,
             ),
-            metadata=(('request-id', request_id), ('session-id', session_id)),
+            metadata=setup_metadata(session_id, request_id),
         )
 
     @expose
     async def commit_index(
         self,
-        index_alias: str,
+        index_name: str,
         commit_mode: Optional[str] = None,
         request_id: Optional[str] = None,
         session_id: Optional[str] = None,
@@ -88,7 +93,7 @@ class SummaClient(BaseGrpcClient):
         after.
 
         Args:
-            index_alias: index alias
+            index_name: index name
             commit_mode: Sync | Async
             request_id: request id
             session_id: session id
@@ -97,10 +102,37 @@ class SummaClient(BaseGrpcClient):
         """
         return await self.stubs['index_api'].commit_index(
             index_service_pb.CommitIndexRequest(
-                index_alias=index_alias,
+                index_name=index_name,
                 commit_mode=commit_mode,
             ),
-            metadata=(('request-id', request_id), ('session-id', session_id)),
+            metadata=setup_metadata(session_id, request_id),
+        )
+
+    @expose
+    async def copy_documents(
+        self,
+        source_index_name: str,
+        target_index_name: str,
+        request_id: Optional[str] = None,
+        session_id: Optional[str] = None,
+    ) -> index_service_pb.CopyDocumentsResponse:
+        """
+        Copies all documents from `source` to `target` index
+
+        Args:
+            source_index_name: source index name
+            target_index_name: target index name
+            request_id: request id
+            session_id: session id
+        Returns:
+            Commit scheduling result
+        """
+        return await self.stubs['index_api'].copy_documents(
+            index_service_pb.CopyDocumentsRequest(
+                source_index_name=source_index_name,
+                target_index_name=target_index_name,
+            ),
+            metadata=setup_metadata(session_id, request_id),
         )
 
     @expose
@@ -135,19 +167,20 @@ class SummaClient(BaseGrpcClient):
                 index_name=index_name,
                 topics=topics,
             ),
-            metadata=(('request-id', request_id), ('session-id', session_id)),
+            metadata=setup_metadata(session_id, request_id),
         )
 
     @expose(with_from_file=True)
     async def create_index(
         self,
         index_name: str,
-        index_engine: str,
-        schema: str,
+        schema: Union[str, list],
+        index_engine: dict,
         compression: Optional[Union[str, int]] = None,
         blocksize: Optional[int] = None,
         sort_by_field: Optional[Tuple] = None,
         index_attributes: Optional[Dict] = None,
+        merge_policy: Optional[Dict] = None,
         request_id: Optional[str] = None,
         session_id: Optional[str] = None,
     ) -> index_service_pb.CreateIndexResponse:
@@ -156,11 +189,12 @@ class SummaClient(BaseGrpcClient):
 
         Args:
             index_name: index name
-            index_engine: "File" or "Memory"
+            index_engine: {"file": {}}, {"memory": {}} or {"ipfs": {"chunked_cache_config": {"chunk_size": ..., "cache_size": ...}}}
             schema: Tantivy index schema
             compression: Tantivy index compression
             blocksize: Docstore blocksize
             index_attributes: Various index attributes such as default fields, multi fields, primary keys etc.
+            merge_policy: describes how to select segments for merging, possible values: {"log": {}}, {"temporal": {}}
             request_id: request id
             session_id: session id
             sort_by_field: (field_name, order)
@@ -169,20 +203,54 @@ class SummaClient(BaseGrpcClient):
             compression = index_service_pb.Compression.Value(compression)
         elif isinstance(compression, int):
             compression = index_service_pb.Compression.Name(compression)
+        if isinstance(schema, list):
+            schema = json.dumps(schema)
         return await self.stubs['index_api'].create_index(
             index_service_pb.CreateIndexRequest(
                 index_name=index_name,
-                index_engine=index_engine,
                 schema=schema,
                 compression=compression,
                 blocksize=blocksize,
                 index_attributes=index_attributes,
                 sort_by_field=index_service_pb.SortByField(
                     field=sort_by_field[0],
-                    order=sort_by_field[1],
-                ) if sort_by_field else None
+                    order=sort_by_field[1].capitalize(),
+                ) if sort_by_field else None,
+                merge_policy=index_service_pb.MergePolicy(**(merge_policy or {})),
+                **index_engine
             ),
-            metadata=(('request-id', request_id), ('session-id', session_id)),
+            metadata=setup_metadata(session_id, request_id),
+        )
+
+    @expose(with_from_file=True)
+    async def copy_index(
+        self,
+        source_index_name: str,
+        target_index_name: str,
+        target_index_engine: dict,
+        merge_policy: Optional[Dict] = None,
+        request_id: Optional[str] = None,
+        session_id: Optional[str] = None,
+    ) -> index_service_pb.CreateIndexResponse:
+        """
+        Copies existing index to new engine
+
+        Args:
+            source_index_name: source index name
+            target_index_name: target index name
+            target_index_engine: {"file": {}}, {"memory": {}} or {"ipfs": {"chunked_cache_config": {"chunk_size": ..., "cache_size": ...}}}
+            merge_policy: describes how to select segments for merging, possible values: {"log": {}}, {"temporal": {}}
+            request_id: request id
+            session_id: session id
+        """
+        return await self.stubs['index_api'].copy_index(
+            index_service_pb.CopyIndexRequest(
+                source_index_name=source_index_name,
+                target_index_name=target_index_name,
+                merge_policy=index_service_pb.MergePolicy(**(merge_policy or {})),
+                **target_index_engine,
+            ),
+            metadata=setup_metadata(session_id, request_id),
         )
 
     @expose
@@ -202,14 +270,14 @@ class SummaClient(BaseGrpcClient):
         """
         return await self.stubs['consumer_api'].delete_consumer(
             consumer_service_pb.DeleteConsumerRequest(consumer_name=consumer_name),
-            metadata=(('request-id', request_id), ('session-id', session_id)),
+            metadata=setup_metadata(session_id, request_id),
         )
 
     @expose
-    async def delete_document(
+    async def delete_documents(
         self,
-        index_alias: str,
-        primary_key: index_service_pb.PrimaryKey,
+        index_name: str,
+        query: query_pb.Query,
         request_id: Optional[str] = None,
         session_id: Optional[str] = None,
     ) -> index_service_pb.IndexDocumentResponse:
@@ -217,17 +285,17 @@ class SummaClient(BaseGrpcClient):
         Delete document with primary key
 
         Args:
-            index_alias: index alias
-            primary_key: bytes
+            index_name: index name
+            query: query used for retrieving documents for deletion
             request_id:
             session_id:
         """
-        return await self.stubs['index_api'].delete_document(
-            index_service_pb.DeleteDocumentRequest(
-                index_alias=index_alias,
-                primary_key=primary_key,
+        return await self.stubs['index_api'].delete_documents(
+            index_service_pb.DeleteDocumentsRequest(
+                index_name=index_name,
+                query=query,
             ),
-            metadata=(('request-id', request_id), ('session-id', session_id)),
+            metadata=setup_metadata(session_id, request_id),
         )
 
     @expose
@@ -247,7 +315,7 @@ class SummaClient(BaseGrpcClient):
         """
         return await self.stubs['index_api'].delete_index(
             index_service_pb.DeleteIndexRequest(index_name=index_name),
-            metadata=(('request-id', request_id), ('session-id', session_id)),
+            metadata=setup_metadata(session_id, request_id),
         )
 
     @expose
@@ -269,7 +337,7 @@ class SummaClient(BaseGrpcClient):
         """
         return await self.stubs['consumer_api'].get_consumer(
             consumer_service_pb.GetConsumerRequest(consumer_name=consumer_name),
-            metadata=(('request-id', request_id), ('session-id', session_id)),
+            metadata=setup_metadata(session_id, request_id),
         )
 
     @expose
@@ -289,13 +357,13 @@ class SummaClient(BaseGrpcClient):
         """
         return await self.stubs['consumer_api'].get_consumers(
             consumer_service_pb.GetConsumersRequest(),
-            metadata=(('request-id', request_id), ('session-id', session_id)),
+            metadata=setup_metadata(session_id, request_id),
         )
 
     @expose
     async def get_index(
         self,
-        index_alias: str,
+        index_name: str,
         request_id: Optional[str] = None,
         session_id: Optional[str] = None,
     ) -> index_service_pb.GetIndexResponse:
@@ -303,15 +371,15 @@ class SummaClient(BaseGrpcClient):
         Index metadata
 
         Args:
-            index_alias: index alias
+            index_name: index name
             request_id: request id
             session_id: session id
         Returns:
             Index description
         """
         return await self.stubs['index_api'].get_index(
-            index_service_pb.GetIndexRequest(index_alias=index_alias),
-            metadata=(('request-id', request_id), ('session-id', session_id)),
+            index_service_pb.GetIndexRequest(index_name=index_name),
+            metadata=setup_metadata(session_id, request_id),
         )
 
     @expose
@@ -331,7 +399,7 @@ class SummaClient(BaseGrpcClient):
         """
         return await self.stubs['index_api'].get_indices(
             index_service_pb.GetIndicesRequest(),
-            metadata=(('request-id', request_id), ('session-id', session_id)),
+            metadata=setup_metadata(session_id, request_id),
         )
 
     @expose
@@ -351,13 +419,34 @@ class SummaClient(BaseGrpcClient):
         """
         return await self.stubs['index_api'].get_indices_aliases(
             index_service_pb.GetIndicesAliasesRequest(),
-            metadata=(('request-id', request_id), ('session-id', session_id)),
+            metadata=setup_metadata(session_id, request_id),
         )
 
+    @expose
+    async def documents(
+        self,
+        index_name: str,
+        request_id: Optional[str] = None,
+        session_id: Optional[str] = None,
+    ) -> AsyncIterator[str]:
+        """
+        Retrieve all documents from the index
+
+        Args:
+            index_name: index name
+            request_id: request id
+            session_id: session id
+        """
+        async for document in self.stubs['index_api'].documents(
+            index_service_pb.DocumentsRequest(index_name=index_name),
+            metadata=setup_metadata(session_id, request_id),
+        ):
+            yield document.document
+
     @expose
     async def index_document_stream(
         self,
-        index_alias: str,
+        index_name: str,
         documents: Union[Iterable[str], str] = None,
         bulk_size: int = 1000,
         request_id: Optional[str] = None,
@@ -367,7 +456,7 @@ class SummaClient(BaseGrpcClient):
         Index documents bulky
 
         Args:
-            index_alias: index alias
+            index_name: index name
             documents: list of bytes
             bulk_size: document portion size to send
             request_id: request id
@@ -385,25 +474,25 @@ class SummaClient(BaseGrpcClient):
                 documents_portion.append(document)
                 if len(documents_portion) > bulk_size:
                     yield index_service_pb.IndexDocumentStreamRequest(
-                        index_alias=index_alias,
+                        index_name=index_name,
                         documents=documents_portion,
                     )
                     documents_portion = []
             if documents_portion:
                 yield index_service_pb.IndexDocumentStreamRequest(
-                    index_alias=index_alias,
+                    index_name=index_name,
                     documents=documents_portion,
                 )
 
         return await self.stubs['index_api'].index_document_stream(
             documents_portion_iter(),
-            metadata=(('request-id', request_id), ('session-id', session_id)),
+            metadata=setup_metadata(session_id, request_id),
         )
 
     @expose
     async def index_document(
         self,
-        index_alias: str,
+        index_name: str,
         document: Union[dict, bytes, str],
         request_id: Optional[str] = None,
         session_id: Optional[str] = None,
@@ -412,25 +501,23 @@ class SummaClient(BaseGrpcClient):
         Index document
 
         Args:
-            index_alias: index alias
+            index_name: index name
             document: bytes
             request_id: request id
             session_id: session id
         """
         return await self.stubs['index_api'].index_document(
             index_service_pb.IndexDocumentRequest(
-                index_alias=index_alias,
+                index_name=index_name,
                 document=json.dumps(document) if isinstance(document, dict) else document,
             ),
-            metadata=(('request-id', request_id), ('session-id', session_id)),
+            metadata=setup_metadata(session_id, request_id),
         )
 
     @expose
     async def search(
         self,
-        index_alias: str,
-        query: dict,
-        collectors: Union[dict, List[dict]],
+        index_queries: List[dict],
         tags: Optional[Dict[str, str]] = None,
         ignore_not_found: bool = False,
         request_id: Optional[str] = None,
@@ -439,32 +526,23 @@ class SummaClient(BaseGrpcClient):
         """Send search request. `Query` object can be created manually or by using `aiosumma.parser` module.
 
         Args:
-            index_alias: index alias
-            query: parsed `Query`
-            collectors: query_pb.Collector list
+            index_queries: index queries
             tags: extra dict for logging purposes
             ignore_not_found: do not raise `StatusCode.NOT_FOUND` and return empty SearchResponse
             request_id: request id
             session_id: session id
         """
-        if isinstance(collectors, (Dict, query_pb.Collector)):
-            collectors = [collectors]
-
         try:
-            search_request = search_service_pb.SearchRequest(
-                index_alias=index_alias,
-                query=query,
-                tags=tags,
-            )
-            for collector in collectors:
-                if isinstance(collector, Dict):
-                    dict_collector = collector
-                    collector = query_pb.Collector()
-                    ParseDict(dict_collector, collector)
-                search_request.collectors.append(collector)
+            search_request = search_service_pb.SearchRequest(tags=tags)
+            for index_query in index_queries:
+                if isinstance(index_query, Dict):
+                    dict_index_query = index_query
+                    index_query = search_service_pb.IndexQuery()
+                    ParseDict(dict_index_query, index_query)
+                search_request.index_queries.append(index_query)
             return await self.stubs['search_api'].search(
                 search_request,
-                metadata=(('request-id', request_id), ('session-id', session_id)),
+                metadata=setup_metadata(session_id, request_id),
             )
         except AioRpcError as e:
             if ignore_not_found and e.code() == StatusCode.NOT_FOUND:
@@ -474,7 +552,7 @@ class SummaClient(BaseGrpcClient):
     @expose
     async def merge_segments(
         self,
-        index_alias: str,
+        index_name: str,
         segment_ids: List[str],
         request_id: Optional[str] = None,
         session_id: Optional[str] = None,
@@ -483,38 +561,14 @@ class SummaClient(BaseGrpcClient):
         Merge a list of segments into a single one
 
         Args:
-            index_alias: index alias
+            index_name: index name
             segment_ids: segment ids
             request_id: request id
             session_id: session id
         """
         return await self.stubs['index_api'].merge_segments(
-            index_service_pb.MergeSegmentsRequest(index_alias=index_alias, segment_ids=segment_ids),
-            metadata=(('request-id', request_id), ('session-id', session_id)),
-        )
-
-    @expose
-    async def publish_index(
-        self,
-        index_alias: str,
-        payload: Optional[str] = None,
-        request_id: Optional[str] = None,
-        session_id: Optional[str] = None,
-    ) -> beacon_service_pb.PublishIndexResponse:
-        """
-        Publish index into IPFS, require `ipfs` section configured in config
-
-        Args:
-            index_alias: index alias
-            payload: extra data stored with publish commit
-            request_id: request id
-            session_id: session id
-        """
-        if isinstance(payload, dict):
-            payload = json.dumps(payload)
-        return await self.stubs['beacon_api'].publish_index(
-            beacon_service_pb.PublishIndexRequest(index_alias=index_alias, payload=payload),
-            metadata=(('request-id', request_id), ('session-id', session_id)),
+            index_service_pb.MergeSegmentsRequest(index_name=index_name, segment_ids=segment_ids),
+            metadata=setup_metadata(session_id, request_id),
         )
 
     @expose
@@ -536,13 +590,13 @@ class SummaClient(BaseGrpcClient):
         """
         return await self.stubs['index_api'].set_index_alias(
             index_service_pb.SetIndexAliasRequest(index_alias=index_alias, index_name=index_name),
-            metadata=(('request-id', request_id), ('session-id', session_id)),
+            metadata=setup_metadata(session_id, request_id),
         )
 
     @expose
     async def vacuum_index(
         self,
-        index_alias: str,
+        index_name: str,
         request_id: Optional[str] = None,
         session_id: Optional[str] = None,
     ) -> index_service_pb.VacuumIndexResponse:
@@ -550,19 +604,41 @@ class SummaClient(BaseGrpcClient):
         Vacuuming index. It cleans every segment from deleted documents, one by one.
 
         Args:
-            index_alias: index alias
+            index_name: index name
             request_id: request id
             session_id: session id
         """
         return await self.stubs['index_api'].vacuum_index(
-            index_service_pb.VacuumIndexRequest(index_alias=index_alias),
-            metadata=(('request-id', request_id), ('session-id', session_id)),
+            index_service_pb.VacuumIndexRequest(index_name=index_name),
+            metadata=setup_metadata(session_id, request_id),
+        )
+
+    @expose
+    async def warmup_index(
+        self,
+        index_name: str,
+        is_full: bool = False,
+        request_id: Optional[str] = None,
+        session_id: Optional[str] = None,
+    ) -> index_service_pb.VacuumIndexResponse:
+        """
+        Warm index up. It loads all hot parts or index into memory and makes further first queries to the index faster.
+
+        Args:
+            index_name: index name
+            is_full: should full or partial warm up be done
+            request_id: request id
+            session_id: session id
+        """
+        return await self.stubs['index_api'].warmup_index(
+            index_service_pb.WarmupIndexRequest(index_name=index_name, is_full=is_full),
+            metadata=setup_metadata(session_id, request_id),
         )
 
     @expose
     async def get_top_terms(
         self,
-        index_alias: str,
+        index_name: str,
         field_name: str,
         top_k: int,
         request_id: Optional[str] = None,
@@ -572,7 +648,7 @@ class SummaClient(BaseGrpcClient):
         Get top terms by the number for the index
 
         Args:
-            index_alias: index alias
+            index_name: index name
             field_name: field name
             top_k: extract top-K terms
             request_id: request id
@@ -580,9 +656,9 @@ class SummaClient(BaseGrpcClient):
         """
         return await self.stubs['reflection_api'].get_top_terms(
             reflection_service_pb.GetTopTermsRequest(
-                index_alias=index_alias,
+                index_name=index_name,
                 field_name=field_name,
                 top_k=top_k,
             ),
-            metadata=(('request-id', request_id), ('session-id', session_id)),
+            metadata=setup_metadata(session_id, request_id),
         )
